# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A package for solving and working with sliding tile puzzles.

All submodule code is imported into the parent namespace, so there is no need
to explicitly refer to submodules.

Examples:
    >>> from slidingpuzzle import *
    >>> b = new_board(3, 3)
    >>> print_board(b)
    1 2 3
    4 5 6
    7 8
    >>> shuffle_board(b)
    >>> print_board(b)
    4 5
    6 7 2
    1 8 3
    >>> r = search(b)
    >>> print_result(r)
    solution_len=20, generated=144704, expanded=103615, unvisited=41090, visited=54466
    >>> r = search(b, algorithm="greedy", heuristic=manhattan_distance)
    >>> print_result(r)
    solution_len=36, generated=409, expanded=299, unvisited=111, visited=153
"""

from slidingpuzzle.heuristics import (
    euclidean_distance,
    hamming_distance,
    manhattan_distance,
    random_distance,
)

from slidingpuzzle.slidingpuzzle import (
    EMPTY_TILE,
    A_STAR,
    BEAM,
    BFS,
    DFS,
    GREEDY,
    IDA_STAR,
    IDDFS,
    ALGORITHMS,
    apply_move,
    compare_heuristics,
    count_inversions,
    freeze_board,
    get_empty_yx,
    get_yx,
    get_next_moves,
    get_next_states,
    is_solvable,
    new_board,
    print_board,
    search,
    shuffle_board,
    shuffle_board_slow,
    solution_as_tiles,
    swap_tiles,
)

from slidingpuzzle.state import SearchResult, State



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module used to evaluate model performance vs. other heuristics.
"""

import torch
import torch.utils


def evaluate(
    model: torch.nn.Module, criterion, dataset: torch.utils.data.Dataset, device
) -> float:
    loss = 0.0
    dataloader = torch.utils.data.DataLoader(dataset)
    model.eval()
    with torch.no_grad():
        for batch, expected in iter(dataloader):
            predicted = model(batch.to(device))
            expected = expected.to(device)
            loss += criterion(predicted, expected).item()
    return loss / len(dataset)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for training a new sliding tile puzzle model for computing heuristic value.

The models compute h(board) -> [0, 1], where
"""

import math

import tensorboard
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
import tqdm

import slidingpuzzle.nn.paths as paths
from slidingpuzzle.nn.dataset import load_dataset
from slidingpuzzle.nn.eval import evaluate


def load_checkpoint(
    model: nn.Module, optimizer: optim.Optimizer = None, epoch: int = None
) -> int:
    """
    Loads a model and optimizer state from the specified epoch.
    If no epoch provided, latest trained model is used.

    Note:
        The board size is extracted from ``model.h`` and ``model.w``,
        which are expected to be present.

    Returns:
        The epoch number found in the checkpoint.
    """
    checkpoint_path = paths.get_checkpoint_path(model.h, model.w, epoch)
    try:
        checkpoint = torch.load(str(checkpoint_path))
        model.load_state_dict(checkpoint["model_state_dict"])
        if optimizer is not None:
            optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        return checkpoint["epoch"]
    except FileNotFoundError:
        print("No model checkpoint found.")
        return 0


def save_checkpoint(model: nn.Module, optimizer: optim.Optimizer, epoch: int) -> None:
    state = {
        "model_state_dict": model.state_dict(),
        "optimizer_state_dict": optimizer.state_dict(),
        "epoch": epoch,
        "board_height": model.h,
        "board_width": model.w,
    }
    # save "latest" model along with epoch labeled checkpoint
    torch.save(state, str(paths.get_checkpoint_path(model.h, model.w)))
    torch.save(state, str(paths.get_checkpoint_path(model.h, model.w, epoch)))


def launch_tensorboard(dirname: str) -> str:
    """
    Launches tensorboard in the specified directory.

    Returns:
        The launch URL.
    """
    tb = tensorboard.program.TensorBoard()
    tb.configure(argv=[None, "--logdir", dirname])
    return tb.launch()


def train(
    model: nn.Module,
    num_examples: int = 10_000,
    train_fraction: float = 0.95,
    num_epochs: int = 50_000,
    batch_size: int = 64,
    device: str = None,
    dataset: torch.utils.data.Dataset = None,
    tensorboard_dir: str = paths.TENSORBOARD_DIR,
) -> nn.Module:
    """
    Trains a model for ``num_epochs``. If no prior model is found, a new one is
    created. If a prior model is found, training will be resumed. If no dataset is
    found, one will be built. Will automatically launch tensorboard and checkpoint
    periodically. The train/test split is done randomly, but in a reproducible way
    so that training may be paused and resumed without leaking test data.

    Note:
        If you change your dataaset between runs, the train/test split will no longer
        be consistent with original training, so training will need to be restarted
        to prevent leaking test data.

    Args:
        h: Height of board to train the model for
        w: Width of the board to train the model for
        num_examples: Total number of examples to use for training / test. Ignored
            if ``dataset`` is provided.
        num_epochs: Total number of epochs model should be trained for
        device: Device to train on. Default will autodetect CUDA or use CPU.
        batch_size: Batch size to use for training
        dataset: A custom dataset to use.
        tensorboard_dir: The root tensorboard dir for logs. Default is "tensorboard".

    Returns:
        The trained model. May be loaded on GPU if GPU is available. The model is also
        saved to disk before returning in the checkpoints directory.
    """
    h, w = model.h, model.w

    # prepare tensorboard to record training
    url = launch_tensorboard(tensorboard_dir)
    print(f"tensorboard launched: {url}")
    log_dir = paths.get_log_dir(tensorboard_dir, h, w)
    comment = f"EXAMPLES_{num_examples}_BS_{batch_size}"
    writer = SummaryWriter(log_dir, comment=comment)

    # prepare dataset
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    dataset = load_dataset(h, w, num_examples) if dataset is None else dataset
    train_size = math.floor(len(dataset) * train_fraction)
    test_size = len(dataset) - train_size
    train_dataset, test_dataset = torch.utils.data.random_split(
        dataset, [train_size, test_size], generator=torch.Generator().manual_seed(0)
    )

    # prepare model
    model.to(device)
    optimizer = optim.SGD(model.parameters(), lr=0.0001, momentum=0.9)
    epoch = load_checkpoint(model, optimizer)
    criterion = nn.MSELoss()

    print(f"initial epoch={epoch}, batch_size={batch_size}")
    try:
        for epoch in tqdm.tqdm(range(epoch, num_epochs)):
            model.train()  # every epoch b/c eval happens every epoch
            running_loss = 0.0

            for batch, expected in iter(
                DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
            ):
                batch = batch.to(device)
                expected = expected.to(device)
                optimizer.zero_grad()
                outputs = model(batch)
                loss = criterion(outputs, expected)
                loss.backward()
                optimizer.step()
                running_loss += loss.item()

            running_loss /= train_size / batch_size
            writer.add_scalar("training_loss", running_loss, epoch)
            test_loss = evaluate(model, criterion, test_dataset, device)
            writer.add_scalar("test_loss", test_loss, epoch)

            if epoch % 100 == 99:
                print(
                    f"training_loss: {running_loss:0.5f}, test_loss: {test_loss:0.5f}"
                )
                save_checkpoint(model, optimizer, epoch)
                writer.flush()

            running_loss = 0.0
    except KeyboardInterrupt:
        print("training interrupted")

    # save final state
    save_checkpoint(model, optimizer, epoch)

    return model



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines neural network-guided heuristics.
"""

import slidingpuzzle.nn.model as model
import slidingpuzzle.nn.paths as paths


MODEL_HEURISTICS = {}


def get_heuristic_key(h: int, w: int, version: str):
    board_size_str = paths.get_board_size_str(h, w)
    return f"{board_size_str}_{version}"


def set_heuristic(h, w, version, heuristic):
    key = get_heuristic_key(h, w, version)
    MODEL_HEURISTICS[key] = heuristic


def get_heuristic(h, w, version):
    key = get_heuristic_key(h, w, version)
    heuristic = MODEL_HEURISTICS.get(key, None)
    if heuristic is None:
        model.load_model(h, w, version)
        heuristic = MODEL_HEURISTICS[key]
    return heuristic


def v1_distance(board: tuple[list[int], ...]) -> float:
    h, w = len(board), len(board[0])
    heuristic = get_heuristic(h, w, model.VERSION_1)
    return heuristic(board)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for creating, saving, and loading board datasets.
"""

import json
import math

import torch
import torch.utils
import tqdm

from slidingpuzzle.slidingpuzzle import (
    apply_move,
    freeze_board,
    new_board,
    search,
    shuffle_board,
    visit,
)
import slidingpuzzle.nn.paths as paths


class SlidingPuzzleDataset(torch.utils.data.Dataset):
    def __init__(self, examples) -> None:
        super().__init__()
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, idx):
        x, y = self.examples[idx]
        return (
            torch.tensor(x, dtype=torch.float32),
            torch.tensor([y], dtype=torch.float32),
        )


def make_examples(h, w, num_examples, prior_examples=None) -> list[tuple]:
    """
    Constructs a list of training examples, which are tuples of:
        (board, num_moves_to_goal)

    Args:
        h: Height of board.
        w: Width of board.
        num_examples: Number of examples to produce.
        prior_examples: Previous examples that we should avoid reproducing.

    Returns:
        The list of training examples.
    """
    examples = [] if prior_examples is None else list(prior_examples)

    if prior_examples is None:
        visited = set()
    else:
        visited = set(freeze_board(example[0]) for example in prior_examples)

    # if the board is large, we need weighted a* to obtain solutions this century
    weight = max(1, math.floor(math.sqrt(h * w) - 2))

    pbar = tqdm.tqdm(total=num_examples)
    while len(examples) < num_examples:
        board = shuffle_board(new_board(h, w))
        if visit(visited, board):
            continue

        # find a path to use as an accurate training reference
        result = search(board, "a*", weight=weight)

        # we can use all intermediate boards as examples
        while len(examples) < num_examples:
            distance = len(result.solution)
            examples.append((freeze_board(board), distance))
            pbar.update(1)
            if not len(result.solution):
                break
            move = result.solution.pop(0)
            apply_move(board, move)
            if visit(visited, board):
                break
    pbar.close()

    return examples


def load_examples(h: int, w: int, examples_file: str = None) -> list:
    """
    Loads examples from a JSON file.
    """
    if examples_file is None:
        examples_file = paths.get_examples_path(h, w)
    with open(examples_file, "rt") as fp:
        return json.load(fp)


def save_examples(h: int, w: int, examples, examples_file: str = None) -> None:
    """
    Save a list of examples to disk as JSON.
    """
    if examples_file is None:
        examples_file = paths.get_examples_path(h, w)
    with open(examples_file, "wt") as fp:
        json.dump(examples, fp)


def load_dataset(
    h: int,
    w: int,
    num_examples: int,
    examples_file=None,
) -> torch.utils.data.Dataset:
    """
    Loads examples, constructs a SlidingPuzzleDataset from them, and returns it.
    If there is a mismatch between the requested num_examples and the provided
    dataset, examples may be truncated or new examples may be constructed and
    saved to the dataset.

    Args:
        h: The height of the board to locate a dataset for
        w: The width of the board to locate a dataset for
        num_examples: The total number of examples desired. If there are too many,
            examples will be truncated. If there are too few, new examples will be
            constructed.
        examples_file: The name of the examples file to save or load.

    Returns:
        A dataset for the requested puzzle size.
    """
    print("Loading dataset...")
    try:
        examples = load_examples(h, w, examples_file)
        print(f"Dataset loaded with {len(examples)} examples.")
    except FileNotFoundError:
        print("No dataset found.")
        examples = []

    if len(examples) > num_examples:
        examples = examples[:num_examples]
    if len(examples) < num_examples:
        new_examples = num_examples - len(examples)
        print(f"Constructing {new_examples} new examples...")
        new_examples = make_examples(h, w, new_examples)
        examples.append(new_examples)
        save_examples(h, w, examples, examples_file)

    return SlidingPuzzleDataset(examples)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A collection of functions for working with sliding tile puzzle boards.
"""

from typing import Callable

import collections
import heapq
import random
import sys

from slidingpuzzle.heuristics import manhattan_distance
from slidingpuzzle.state import SearchResult, State


# prevent typos
EMPTY_TILE = 0
A_STAR = "a*"
BEAM = "beam"
BFS = "bfs"
DFS = "dfs"
GREEDY = "greedy"
IDA_STAR = "ida*"
IDDFS = "iddfs"
ALGORITHMS = (
    A_STAR,
    BEAM,
    BFS,
    DFS,
    GREEDY,
    IDA_STAR,
    IDDFS,
)


def new_board(h: int, w: int) -> tuple[list[int], ...]:
    """
    Create a new board in the default solved state.

    Args:
        h: Height of the board.
        w: Width of the board.

    Returns:
        The new board.
    """
    board = tuple([(y * w) + x + 1 for x in range(w)] for y in range(h))
    board[-1][-1] = 0
    return board


def freeze_board(board: tuple[list[int], ...]) -> tuple[tuple[int, ...], ...]:
    """
    Obtain a frozen copy of the board that is hashable.

    Args:
        board: The board to freeze.

    Returns:
        A tuple of tuple[int].
    """
    return tuple(tuple(row) for row in board)


def print_board(board: tuple[list[int], ...], file=sys.stdout) -> None:
    """
    Convienance function for printing a formatted board.

    Args:
        board: The board to print.
        file: The target output file. Defaults to stdout.
    """
    board_size = len(board) * len(board[0])
    # the longest str we need to print is the largest tile number
    max_width = len(str(board_size - 1))
    for row in board:
        for tile in row:
            if tile == EMPTY_TILE:
                print(" " * max_width, end=" ", file=file)
            else:
                print(str(tile).ljust(max_width), end=" ", file=file)
        print(file=file)


def get_yx(board: tuple[list[int], ...], tile: int) -> tuple[int, int]:
    """
    Given a tile number, find the (y, x)-coord on the board.

    Args:
        board: The puzzle board.
        move: The

    Returns:
        The tile position or None if
    """
    for y, row in enumerate(board):
        for x in range(len(row)):
            if board[y][x] == tile:
                return y, x
    raise ValueError(f'There is no tile "{tile}" on the board.')


def get_empty_yx(board: tuple[list[int], ...]) -> tuple[int, int]:
    """
    Locate the empty tile's (y, x)-coord.

    Args:
        board: The puzzle board.

    Returns:
        The (y, x)-coord of the empty tile.
    """
    return get_yx(board, EMPTY_TILE)


def get_next_moves(
    board: tuple[list[int], ...],
    empty_pos: tuple[int, int] = None,
) -> list[tuple[int, int]]:
    """
    Return a list of all possible moves.

    Args:
        board: The current puzzle board.
        empty_pos: The position of the empty tile.
            If not provided, it will be located automatically.

    Returns:
        A list of (y, x)-coords that are tile positions capable of
        swapping with the empty tile.
    """
    if empty_pos is None:
        empty_pos = get_empty_yx(board)
    y, x = empty_pos
    moves = []
    for dy, dx in ((0, -1), (0, 1), (-1, 0), (1, 0)):
        if 0 <= y + dy < len(board) and 0 <= x + dx < len(board[0]):
            moves.append((y + dy, x + dx))
    return moves


def swap_tiles(
    board: tuple[list[int], ...], pos1: tuple[int, int], pos2: tuple[int, int]
) -> None:
    """
    Mutates the board by swapping a pair of tiles.

    Args:
        board: The board to modify.
        pos1: The first tile position.
        pos2: The second tile position.

    Return:
        The modified board, for conveience chaining calls.
    """
    y1, x1 = pos1
    y2, x2 = pos2
    board[y1][x1], board[y2][x2] = board[y2][x2], board[y1][x1]
    return board


def count_inversions(board: tuple[list[int], ...]) -> int:
    """
    From each tile, count the number of tiles that are out of place.
    Returns the sum of all counts. See :func:`is_solvable`.

    Args:
        board: The puzzle board.

    Returns:
        The count of inversions.
    """
    h, w = len(board), len(board[0])
    board_size = h * w
    inversions = 0
    for tile1 in range(board_size):
        for tile2 in range(tile1 + 1, board_size):
            t1 = board[tile1 // w][tile1 % w]
            t2 = board[tile2 // w][tile2 % w]
            if EMPTY_TILE in (t1, t2):
                continue
            if t2 < t1:
                inversions += 1
    return inversions


def apply_move(
    board: tuple[list[int], ...],
    move: tuple[int, int] | int,
    empty_pos: tuple[int, int] = None,
) -> tuple[list[int], ...]:
    """
    Applies a move to the board in place.

    Args:
        board: The puzzle board.
        move: The move to apply. Can be a (y, x)-coord or a tile int.
        empty_pos: The position of the empty tile. Will be found if not provided.

    Returns:
        The modified board, for convience chaining calls.
    """
    if isinstance(move, int):
        move = get_yx(move)
    if empty_pos is None:
        empty_pos = get_empty_yx(board)
    return swap_tiles(board, move, empty_pos)


def is_solvable(board: tuple[list[int], ...]) -> bool:
    """
    Determines if it is possible to solve this board.

    Note:
        The algorithm counts `inversions`_ to determine solvability.
        The "standard" algorithm has been modified here to support
        non-square board sizes.

    Args:
        board: The puzzle board.

    Returns:
        bool: True if the board is solvable, False otherwise.

    .. _inversions:
        https://www.cs.princeton.edu/courses/archive/spring21/cos226/assignments/8puzzle/specification.php
    """
    inversions = count_inversions(board)
    h, w = len(board), len(board[0])
    if w % 2 == 0:
        y, _ = get_empty_yx(board)
        if h % 2 == 0:
            if (inversions + y) % 2 != 0:
                return True
        else:
            if (inversions + y) % 2 == 0:
                return True
    elif inversions % 2 == 0:
        return True
    return False


def shuffle_board(board: tuple[list[int], ...]) -> tuple[list[int], ...]:
    """
    Shuffles a board (in place) and validates that the result is solvable.

    Args:
        board: The board to shuffle.

    Returns:
        The same board for chaining convience.
    """
    h, w = len(board), len(board[0])
    while True:
        # first shuffle the board
        for y in range(h):
            for x in range(w):
                pos1 = y, x
                pos2 = random.randrange(h), random.randrange(w)
                swap_tiles(board, pos1, pos2)

        if is_solvable(board):
            break
    return board


def shuffle_board_slow(
    board: tuple[list[int], ...], num_moves: int = None, moves: list = None
) -> tuple[list[int], ...]:
    """
    Shuffles a board in place by making random legal moves.
    Each move is first checked to avoid repeated states, although
    this does not guarantee the

    Args:
        board: The board to shuffle.
        num_moves (int): Number of random moves to make.
            If ``None``, ``(h * w) ** 2`` will be used.
        moves: If a list is provided, the moves made will be appended.

    Returns:
        The same board for chaining convience.
    """
    h, w = len(board), len(board[0])
    if num_moves is None:
        num_moves = (h * w) ** 2
    empty_pos = get_empty_yx(board)
    visited = set()
    for _ in range(num_moves):
        next_moves = get_next_moves(board, empty_pos)
        random.shuffle(next_moves)
        new_move = False  # tracks if this is a new move
        # try to find a new move
        while next_moves:
            next_move = next_moves.pop()
            swap_tiles(board, empty_pos, next_move)
            # have we been here before?
            if visit(visited, board):
                swap_tiles(board, empty_pos, next_move)  # undo move
            else:
                new_move = True
                break

        # if we couldn't find a new move, we're done
        if not new_move:
            return board

        # if we're tracking which moves we made, record this move
        if moves is not None:
            moves.append(next_move)

        # update empty_pos
        empty_pos = next_move
    return board


def get_next_states(state: State) -> list[State]:
    """
    Creates a list of next viable states, given the current state.
    """
    moves = get_next_moves(state.board, state.empty_pos)
    next_states = []
    for move in moves:
        # construct new altered board
        next_board = tuple(row.copy() for row in state.board)
        swap_tiles(next_board, state.empty_pos, move)
        # record history
        next_history = state.history.copy()
        next_history.append(move)
        # after moving, the move_pos is now the empty_pos
        next_state = State(next_board, move, next_history)
        next_states.append(next_state)
    return next_states


def visit(
    visited: set[tuple[tuple[int, ...], ...]], board: tuple[list[int], ...]
) -> bool:
    """
    Helper to check if this state already exists. Otherwise, record it.
    Returns True if we have already been here, False otherwise.

    Args:
        visited: Set of boards already seen.
        board: The current board.

    Returns:
        True if we've been here before.
    """
    frozen_board = freeze_board(board)
    if frozen_board in visited:
        return True
    visited.add(frozen_board)
    return False


def search(
    board: tuple[list[int], ...],
    algorithm: str = A_STAR,
    heuristic: Callable[[tuple[list[int], ...]], int | float] = None,
    **kwargs,
) -> SearchResult:
    """
    Searches for a set of moves that take the provided board state to the
    solved state.

    Requested algorithm may be one of:
        "a*", "beam", "bfs", "dfs", "greedy", "ida*", "iddfs"

    See :mod:`heuristics` for heuristic functions or provide your own.

    If a heuristic is provided with "bfs", "dfs", or "iddfs", it is only used
    to sort the locally generated nodes on each expansion, but not the entire
    open list. You may also provide a "bound" via the kwargs described below
    to limit the search depth.

    If "greedy" or "a*" is used, the entire open list is sorted. The "greedy"
    algorithm sorts only on the heuristic, while "a*" uses heuristic + length
    of the solution. Default heuristics are listed in the table below.

    Of the provided algorithms, only beam search is incomplete. This means it
    may miss the goal, even thought the board is solvable.

    The algorithms support some additional kwargs that can be used to
    customize their behavior. These are:

    ..  code-block:: python

        a*: {
            "bound": default is float("inf"),
                restricts search to f-values < bound
            "heuristic": default is manhattan_distance
            "weight": default is 1
        }
        beam: {
            "bound": default is float("inf"),
                restricts search to f-values < bound
            "heuristic": default is manhattan_distance
            "width", default is 3
        }
        bfs: {
            "bound": default is float("inf"),
                restricts search to depth < bound
            "heuristic": None
        }
        dfs: {
            "bound": default is float("inf"),
                restricts search to depth < bound
            "heuristic": None
        }
        ida*: {
            "bound", default is manhattan_distance(board),
                restricts search to f-values < bound
            "heuristic": default is manhattan_distance
            "weight": default is 1
        }
        iddfs: {
            "bound", default is manhattan_distance(board),
                restricts search to depth < bound
            "heuristic": None
        }

    Args:
        board: The initial board state to search.
        algorithm (str): The algorithm to use for search.
        heuristic: A function to evaluate the board.
            See :mod:`slidingpuzzle.heuristics`.

    Returns:
        Returns a :class:`SearchResult` containing a list of moves to solve the
        puzzle from the initial state along with some search statistics. Solution
        may be None if no solution was found.
    """

    if not is_solvable(board):
        raise ValueError("The provided board is not solvable.")
    algorithm = algorithm.strip().lower()
    if algorithm not in ALGORITHMS:
        raise ValueError(f'Unknown algorithm: "{algorithm}"')
    if heuristic is None:
        if algorithm in (A_STAR, BEAM, IDA_STAR):
            heuristic = manhattan_distance
        else:
            # if no heuristic is provided, treat all nodes equally
            def heuristic(*_):
                return 0

    # only relevant for some algorithms
    a_star_weight = kwargs.get("weight", 1)
    beam_width = kwargs.get("width", 3)
    if beam_width <= 0:
        raise ValueError("Beam width <= 0 is impossible.")

    if algorithm in (IDA_STAR, IDDFS):
        # initial upper bound for ida* is the minimum number of moves possible
        bound = kwargs.get("bound", manhattan_distance(board))
    else:
        bound = kwargs.get("bound", float("inf"))

    # this is used by iterative deepening algorithms to determine when the
    # search space has been exhausted and/or to set the next iteration bound
    next_bound = float("inf")

    # prepare a func that will be used to evaluate states, depending upon alg.
    if algorithm in (A_STAR, IDA_STAR):

        def evaluate(state):
            return len(state.history) + a_star_weight * heuristic(state.board)

    else:

        def evaluate(state):
            return heuristic(state.board)

    # prepare initial state
    board = tuple(row.copy() for row in board)  # don't modify orig board
    goal = new_board(len(board), len(board[0]))
    empty_pos = get_empty_yx(board)
    expanded = 0
    generated = 0
    initial_state = State(board, empty_pos, [])
    visited = set()  # closed states
    if algorithm in (BEAM, BFS):
        unvisited = collections.deque([initial_state])
    else:
        unvisited = [initial_state]

    def get_next_state() -> State:
        """
        A helper function to properly remove the next state based upon
        the underlying data structure.
        """
        if algorithm in (BEAM, BFS):
            return unvisited.popleft()
        elif algorithm in (A_STAR, GREEDY, IDA_STAR):
            return heapq.heappop(unvisited)
        else:
            return unvisited.pop()

    def add_states(states) -> None:
        """Add provided states to unvisited."""
        if algorithm in (A_STAR, GREEDY, IDA_STAR):
            for s in states:
                heapq.heappush(unvisited, s)
        else:
            unvisited.extend(states)

    while unvisited:
        # some search algorithms need to restart search with an altered state.
        # see code after this inner loop
        while unvisited:
            # breadth-first search removes from the front, all others use the end
            state = get_next_state()
            expanded += 1

            # bounds check (if not using bound, this will always pass)
            if algorithm in (A_STAR, BEAM, IDA_STAR):
                if state.f > bound:
                    next_bound = min(next_bound, state.f)  # used by ida*
                    continue
            else:
                depth = len(state.history)
                if depth > bound:
                    next_bound = depth  # used by iddfs
                    continue

            # most algorithms check for duplicate states
            if algorithm not in (IDA_STAR, IDDFS):
                if visit(visited, state.board):
                    continue

            # goal check
            if state.board == goal:
                return SearchResult(
                    board, generated, expanded, list(unvisited), visited, state.history
                )

            # compute value of next states
            next_states = get_next_states(state)
            for next_state in next_states:
                next_state.f = evaluate(next_state)

            # update the open list
            if algorithm in (A_STAR, GREEDY, IDA_STAR):
                add_states(next_states)
            else:
                # these algorithms sort only local nodes
                next_states.sort(reverse=True)
                # we should only store the last beam_width states during beam search
                if BEAM == algorithm:
                    next_states = next_states[-beam_width:]
                add_states(next_states)
            generated += len(next_states)

        # do we need to restart search?
        if algorithm in (IDA_STAR, IDDFS) and bound != next_bound:
            bound = next_bound
            if IDA_STAR == algorithm:
                next_bound = float("inf")
            unvisited.append(initial_state)

    # if we are here, no solution was found
    return SearchResult(board, generated, expanded, list(unvisited), visited, None)


def solution_as_tiles(
    board: tuple[list[int], ...], solution: list[tuple[int, int]]
) -> list[int]:
    """
    Converts a list of (y, x)-coords indicating moves into tile numbers,
    given a starting board configuration.

    Args:
        board: The initial board we will apply moves to (does not alter board).
        solution: A list of move coordinates in (y, x) form.

    Returns:
        A list of ints, which indicate a sequence of tile numbers to move.
    """
    board = tuple(row.copy() for row in board)
    tiles = []
    empty_pos = get_empty_yx(board)
    for move in solution:
        y, x = move
        tiles.append(board[y][x])
        swap_tiles(board, empty_pos, move)
        empty_pos = move
    return tiles


def compare_heuristics(
    h: int,
    w: int,
    heuristic_a,
    heuristic_b=manhattan_distance,
    num_iters: int = 30,
    algorithm=A_STAR,
    **kwargs,
):
    """
    Runs search on ``num_iters`` random boards, first using A* with the heuristic A,
    heuristic and then again on the same board on heuristic B. Returns the average
    number of nodes expanded for both.

    Returns:
        A tuple containing (avg. expanded A, avg. expanded B),
        where "heuristic" is your provided heuristic.
    """
    avg_a, avg_b = 0, 0
    for _ in range(num_iters):
        board = new_board(h, w)
        shuffle_board(board)
        result = search(board, algorithm=algorithm, heuristic=heuristic_a, **kwargs)
        avg_a += result.expanded
        result = search(board, algorithm=algorithm, heuristic=heuristic_b, **kwargs)
        avg_b += result.expanded
    return avg_a / num_iters, avg_b / num_iters



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import random

import pytest

from slidingpuzzle import *


def test_newboard():
    ground_truth = ([1, 2, 3], [4, 5, 6], [7, 8, 0])
    assert new_board(3, 3) == ground_truth

    ground_truth = ([1, 2], [3, 4], [5, 0])
    assert new_board(3, 2) == ground_truth


def test_freezeboard():
    b = new_board(3, 3)
    s = set()
    frozen = freeze_board(b)
    s.add(frozen)


def test_print_board():
    print_board(new_board(5, 5))


def test_get_yx():
    board = new_board(3, 3)
    tile = 1
    for y, row in enumerate(board):
        for x in range(len(row)):
            if board[y][x] == EMPTY_TILE:
                continue
            assert tile == board[y][x]
            tile += 1


def test_get_empty_yx():
    board = new_board(3, 3)
    assert get_empty_yx(board) == (2, 2)
    assert get_yx(board, EMPTY_TILE) == get_empty_yx(board)

    tmp = board[2][2]
    board[2][2] = board[0][1]
    board[0][1] = tmp
    assert get_empty_yx(board) == (0, 1)


def test_swap_tiles():
    board = new_board(3, 3)
    pos1 = (2, 2)
    pos2 = (1, 1)
    orig_pos1 = board[pos1[0]][pos1[1]]
    orig_pos2 = board[pos2[0]][pos2[1]]
    swap_tiles(board, pos1, pos2)
    assert orig_pos1 == board[pos2[0]][pos2[1]]
    assert orig_pos2 == board[pos1[0]][pos1[1]]


def test_apply_move():
    board = new_board(5, 5)
    pos1 = (2, 2)
    pos2 = get_empty_yx(board)
    orig_pos1 = board[pos1[0]][pos1[1]]
    orig_pos2 = board[pos2[0]][pos2[1]]
    apply_move(board, pos1)
    assert orig_pos1 == board[pos2[0]][pos2[1]]
    assert orig_pos2 == board[pos1[0]][pos1[1]]


@pytest.mark.parametrize("size", [(2, 2), (3, 2), (2, 3), (3, 3)])
def test_shuffle_board(size):
    random.seed(0)
    h, w = size
    board = new_board(h, w)
    shuffle_board(board)
    r = search(board, algorithm="greedy", heuristic=manhattan_distance)
    assert r.solution is not None


@pytest.mark.parametrize("size", [(2, 2), (3, 2), (2, 3), (3, 3)])
def test_shuffle_board_slow(size):
    random.seed(0)
    h, w = size
    board = new_board(h, w)
    shuffle_board_slow(board)
    r = search(board, algorithm="greedy", heuristic=manhattan_distance)
    assert r.solution is not None


def test_get_possible_moves():
    board = new_board(3, 3)
    moves = get_next_moves(board, (2, 2))
    assert moves == [(2, 1), (1, 2)]

    pos1 = (1, 1)
    pos2 = (2, 2)
    swap_tiles(board, pos1, pos2)
    moves = get_next_moves(board, pos1)
    assert len(moves) == 4


def test_get_next_states():
    board = new_board(3, 3)
    state = State(board, (2, 2), [])
    next_states = get_next_states(state)
    assert len(next_states) == 2


def test_hamming():
    board = new_board(5, 3)
    assert hamming_distance(board) == 0

    swap_tiles(board, (0, 1), (0, 0))
    assert hamming_distance(board) == 2


def test_manhattan():
    board = new_board(5, 3)
    assert manhattan_distance(board) == 0

    swap_tiles(board, (1, 2), (0, 0))
    assert manhattan_distance(board) == 6


def test_euclidean():
    board = new_board(5, 3)
    assert euclidean_distance(board) == 0

    swap_tiles(board, (1, 2), (0, 0))
    c = math.sqrt((2 * 1) ** 2 + (2 * 2) ** 2)
    assert euclidean_distance(board) == c


def test_a_star():
    board = new_board(3, 3)
    shuffle_board_slow(board, num_moves=10)
    assert len(search(board).solution) == len(
        search(board, A_STAR, manhattan_distance).solution
    )


@pytest.mark.parametrize("algorithm", [IDA_STAR, IDDFS])
def test_search_slow(algorithm):
    random.seed(0)
    b = new_board(3, 3)
    shuffle_board_slow(b, 10)
    expected_len = len(search(b).solution)
    actual = search(b, algorithm=algorithm, heuristic=manhattan_distance)
    # there should be no history recorded for iterative deepening
    assert len(actual.visited) == 0
    # best solution should be found (using default args)
    actual_len = len(actual.solution)
    assert expected_len == actual_len


@pytest.mark.parametrize("algorithm", [A_STAR, BEAM, BFS, DFS, GREEDY])
@pytest.mark.parametrize(
    "heuristic", [None, euclidean_distance, manhattan_distance, random_distance]
)
def test_search(algorithm, heuristic):
    random.seed(0)
    board = tuple([[5, 2, 4], [3, 0, 1]])
    result = search(
        board,
        algorithm=algorithm,
        heuristic=heuristic,
        kwargs={"weight": 2, "width": 3},
    )
    if BFS == algorithm:
        assert len(result.solution) == 15
    else:
        assert len(result.solution) >= 15
    print(repr(result))
    print(str(result))


def test_search_hard():
    random.seed(0)
    board = tuple([[8, 6, 7], [3, 5, 1], [2, 0, 4]])
    result = search(
        board,
        algorithm=A_STAR,
        heuristic=manhattan_distance,
        kwargs={"weight": 1},
    )
    assert len(result.solution) == 27


def test_heuristic_behavior():
    random.seed(0)

    # ordered from best -> worst
    heuristics = [
        manhattan_distance,
        euclidean_distance,
        hamming_distance,
        random_distance,
    ]

    # we compute avg expanded nodes over multiple runs to confirm that
    # heuristic behavior is in line with expectations
    num_runs = 20
    expanded_avg = [0 for _ in range(len(heuristics))]
    for i, heuristic in enumerate(heuristics):
        for _ in range(num_runs):
            board = new_board(3, 2)  # largest puzzle reasonably solvable with random
            shuffle_board(board)
            result = search(board, algorithm="greedy", heuristic=heuristic)
            expanded_avg[i] += result.expanded
        expanded_avg[i] /= num_runs
    print("expanded_avg:", expanded_avg)

    # manhattan + euclidean are both good contenders, so we don't compare them
    assert expanded_avg[0] < expanded_avg[2]
    assert expanded_avg[1] < expanded_avg[2]
    assert expanded_avg[2] < expanded_avg[3]


def test_solution_as_tiles():
    h, w = 3, 3
    b = new_board(h, w)
    swap_tiles(b, (h - 1, w - 1), (h - 2, w - 1))
    swap_tiles(b, (h - 2, w - 1), (h - 2, w - 2))
    r = search(b)
    assert [5, 6] == solution_as_tiles(b, r.solution)
